import csv
import simplejson as json
import os
import glob
import copy
import time

class BLEProcessor:
   sensor_map={}
   templates={}

   def __init__(self,csv_format=1):
       self.csv_format = csv_format

   def make_subassembly(self,collector):
       instance = {
         collector['subAssembly']['lowercase_name']: {
            "collectors": {
               collector['name']: {
                  "CSV": []
               } 
            }
         } 
       }
       return instance

   def get_templates(self):
        return self.templates

   def add_sensor_mapping(self,sensorAssignments):
        #print("sensorAssignments: ",json.dumps(sensorAssignments,indent=4))
        subassemblies={}
        for sensor in sensorAssignments:
           #print("Sensor: ",json.dumps(sensor,indent=4))
           machineId = sensor['machineId']
           sensor_mac = sensor['sensorId'].split('_')[0].upper()
           base_name= sensor['sensorId'].split('_')[1].lower()
           sensor_key=sensor_mac + ":" + base_name
           if sensor_key not in self.sensor_map:
               self.sensor_map[sensor_key] = {"machineId":machineId}
               #self.make_subassembly(sensor['collectors'])
           #else:
           #    self.make_subassembly(sensor['collectors'])
           collector = sensor['collectors']
           #subassembly_name = collector['subAssembly']['lowercase_name']
           subassembly_name = sensor['subAssemblyInstance']['name']
           #print("subassembly_name: ",subassembly_name)
           if subassembly_name not in subassemblies:
              subassemblies[subassembly_name] = {"collectors":{}}
           subassemblies[subassembly_name]["collectors"][collector['name']] = {"CSV":[]}
           self.sensor_map[sensor_key]['assembly_key'] = subassembly_name
           self.sensor_map[sensor_key]['collector_key'] = collector['name']
        return subassemblies

   def make_header(self,config):
       header={}
       header={
		'pumpNumber': config.get('pumpNumber'),
                'locationName': config['zone']['location']['name'],
                'companyName': config['zone']['location']['company']['name'],
                'assemblyId': config['models']['assemblyId'],
                'machineName': config.get('display_name'),
                "manufacturer": config.get('category'),  # Not sure if that is right.
                "modelName": config['models']['name'],                                  
                "companyId": config['zone']['location']['company']['id'],                                        
                "machineId": config['sensorAssignments'][0]['machineId'],
                "assemblyName": config['models']['assembly']['name'],       
                "zoneId": config.get('zoneId'),
                "gaugeDuration": int(2), 	# Not sure why this is hard coded.   
                "zoneName": config['zone']['name'],  
                "qc_profile": config.get('qc_profile'),  
                "group": config.get('group'),       
                "timestamp": None
       }
       return header

   def create_template_from_config(self,config):
      header=self.make_header(config)
      subassemblies = self.add_sensor_mapping(config['sensorAssignments'])
      template = {
         'header':header,
         'subassemblies':subassemblies
      }
      return template
   
   def add_config_to_template(self,config):
       pass

   def load_machine_file(self,machine_file):
       try:
          config = json.loads(open(machine_file).read())
          template = self.create_template_from_config(config)
          self.templates[template['header']['machineId']] = template
       except Exception as e:
          try:
             print(e)
          finally:
             e = None
             del e


   def load_machines(self,path):
        mp = os.path.join(path,'machine_details_*.json')
        machine_details_file_list_arr = glob.glob(mp)
        for machine_details in machine_details_file_list_arr:
            self.load_machine_file(machine_details)
        #print("Sensor Map: ",json.dumps(self.sensor_map,indent=4))


   def load_csv_format1(self,csvpath):
        file_values={}
        with open(csvpath) as (csvfile):                                              
            readCSV = csv.reader(csvfile, delimiter=',')                              
            for row in readCSV:
                topic=row[1];
                topic_temp_arr = topic.split('/')
                sensor_mac_add = ''.join(topic_temp_arr[2].split(':'))
                CSV=[];
                CSV.append(row[0])
                CSV = CSV + row[3:]
                if not sensor_mac_add in file_values:
                   file_values[sensor_mac_add] = {
                      topic_temp_arr[1]: {
                         "CSV": [CSV]
                      }
                   }
                else:
                   if topic_temp_arr[1] in file_values[sensor_mac_add]:
                      file_values[sensor_mac_add][topic_temp_arr[1]]["CSV"].append(CSV)
                   else:
                      file_values[sensor_mac_add][topic_temp_arr[1]] = {"CSV":[CSV]}
        return file_values

   def csv_parse_format2(self,row,file_values):
       sensor_mac_add=row[1].replace(":","")
       sensor_name=row[4]
       CSV=[];
       CSV.append(row[0])
       CSV = CSV + row[6:]
       if not sensor_mac_add in file_values:
           file_values[sensor_mac_add] = {
              sensor_name: {
                 "CSV": [CSV]
              }
           }
       else:
           if sensor_name in file_values[sensor_mac_add]:
              file_values[sensor_mac_add][sensor_name]["CSV"].append(CSV)
           else:
              file_values[sensor_mac_add][sensor_name] = {"CSV":[CSV]}
   
   def load_csv_format2(self,csvpath):
        file_values={}
        with open(csvpath) as (csvfile):                                              
            readCSV = csv.reader(csvfile, delimiter=',')                              
            for row in readCSV:
                self.csv_parse_format2(row,file_values)
        return file_values
   
   def load_csv(self,csvpath):
       if self.csv_format == 2:
           return self.load_csv_format2(csvpath)
       return self.load_csv_format1(csvpath)

   def get_mapping(self,key):
      return self.sensor_map.get(key)
  
   def get_input_json_template(self,machine_id):
      if machine_id in self.templates:
          return copy.deepcopy(self.templates[machine_id])
      return None

   def create_input_jsons(self,input_values,timestamp=None):
      input_jsons = {}
      if timestamp is None:
          timestamp = time.time()
      for sensor_mac, values in input_values.items():
          for collect_key, csvs in values.items():
             sensor_key = sensor_mac + ":" + collect_key
             mapping = self.get_mapping(sensor_key)
             #print("Mapping: ",json.dumps(mapping,indent=4))
             if not mapping:

                print("No Mapping for sensor key: ",sensor_key)
                continue

             if mapping['machineId'] not in input_jsons:
                template = self.get_input_json_template(mapping['machineId'])
                if template:
                   template['header']['timestamp'] = timestamp
                   input_jsons[mapping['machineId']] = template
             if mapping['machineId'] in input_jsons:           # Same test again, but now if it is in, then we have a template.
                 #print("INPUT JSON: ",json.dumps(input_jsons[mapping['machineId']],indent=4))
                 input_jsons[mapping['machineId']]['subassemblies'][mapping['assembly_key']]['collectors'][mapping['collector_key']]['CSV'] = csvs['CSV']
      return input_jsons

   def process_ble_file_tojson(self,file):
       basename=os.path.basename(file)
       timestamp = str(basename[:basename.find('.')])
       values=self.load_csv(file)
       #print("VALUES OUT: ",json.dumps(values,indent=4))
       jsons = self.create_input_jsons(values,timestamp=timestamp)
       #print("Input jsons: ",json.dumps(jsons,indent=4))
       return jsons
   
   def process_ble_list_tojson(self,list,timestamp=None):
       values={}
       for line in list:
           row=line.split(",")
           self.csv_parse_format2(row,values)
       #print("VALUES OUT: ",json.dumps(values,indent=4))
       jsons = self.create_input_jsons(values,timestamp=timestamp)
       #print("Input jsons: ",json.dumps(jsons,indent=4))
       return jsons

