"""
[-] Title: PUMP COMPUTATION
[~] Build: v1.0
[+] Author: Biprodeep Roy
[>] Licensed to: MachineSense™
[~] Description: This application retrives the raw data from GPIO Process (Gatttool) and converts it to TELEGRAF format and then sends it to cloud.
"""
import time, os, csv, glob

import simplejson as json
from threading import Thread
import time
#from pumpsense_computation.pump_analytics import main_output
from qc_computation.ac_vac_qc_analytics import main_output_qc
import json, copy, traceback
        
qc_config_file="/tmp/qc/conf/qc_input_configuration.json"

class Pumpsense_Computation:

    def banner(self):
        print('\n                                                    ##########################################################\n                                                    #   [-] Title: PUMP COMPUTATION                          #\n                                                    #   [~] Build: v2.1 (2021)                               #\n                                                    #   [+] Author: Biprodeep Roy (@virtualgh0st), Biswadeep #\n                                                    #   [>] Licensed to: MachineSense™                       #\n                                                    #   [~] Report Bugs: broy@machinesense.com               #\n                                                    ##########################################################\n                    ')

    def __init__(self):
        self.bridge_data_path = '/tmp/qc/ramcsv/'
        self.machine_filepath = '/opt/novatec/data/slave/json_data/machine_details_0.json'
        self.machine_json_dict = {}
        self.input_json_dict = {}
        self.sensor_mac_map_dict = {}
        self.subAssembly_input = {}
        self.subAssembly_input_replicate = {}
        self.batch_memory_queue = []
        self.qc_switch = False
        self.flagStateCount = 0
        self.qc_input_configuration()

    def main(self):
        self.banner()
        self.initialise_input_json()
        for macid in self.input_json_dict:
            self.subAssembly_input[macid] = {}
            self.subAssembly_input[macid]['subassemblies'] = {}
            self.subAssembly_input[macid]['subassemblies'] = self.input_json_dict[macid]['subassemblies']
            del self.input_json_dict[macid]['subassemblies']

        loop_memory_queue_thread = Thread(target=(self.loop_memory_queue))
        loop_memory_queue_thread.start()
        while 1:
            if len(self.batch_memory_queue) > 0:
                for bmq in self.batch_memory_queue:
                    try:
                        timestamp = ''
                        temp_ij = {}
                        for macid in self.input_json_dict:
                            temp_ij[macid] = self.input_json_dict[macid]

                        for macid in bmq:
                            temp_ij[macid]['header']['timestamp'] = int(bmq[macid]['timestamp'])
                            temp_ij[macid]['subassemblies'] = bmq[macid]['subassemblies']
                            try:
                                temp_ij[macid]['header']['qc_profile']['flagStateCount'] = self.flagStateCount
                            except Exception as e:
                                try:
                                    traceback.print_exc()
                                    print(e)
                                    temp_ij[macid]['header']['qc_profile'] = {}
                                    temp_ij[macid]['header']['qc_profile']['flagStateCount'] = self.flagStateCount
                                finally:
                                    e = None
                                    del e
                        for macid in temp_ij:
                            print('[+] QC Flag: ' + str(self.qc_switch))
                            print("INPUT JSON: ",json.dumps(temp_ij[macid],indent=4))
                            temp_ij[macid]['header']['group']='qc'
                            main_output_qc(temp_ij[macid])
                            #if self.qc_switch:
                            #    main_output_qc(temp_ij[macid])
                            #else:
                            #    main_output(temp_ij[macid])

                        self.batch_memory_queue.remove(bmq)
                        time.sleep(2)
                    except Exception as e:
                        try:
                            traceback.print_exc()
                            print(e)
                            self.batch_memory_queue.remove(bmq)
                            time.sleep(2)
                            continue
                        finally:
                            e = None
                            del e

            else:
                time.sleep(5)

    def refetch_machine_details(self):
        self.initialise_input_json()
        for macid in self.input_json_dict:
            self.subAssembly_input[macid] = {}
            self.subAssembly_input[macid]['subassemblies'] = {}
            self.subAssembly_input[macid]['subassemblies'] = self.input_json_dict[macid]['subassemblies']
            del self.input_json_dict[macid]['subassemblies']

    def qc_input_configuration(self):
        global qc_config_file

        tempj = {}
        #with open('/usr/src/conf/qc_input_configuration.json') as (f):
        with open(qc_config_file) as (f):
            tempj = json.load(f)
        if tempj['qc_switch']:
            self.qc_switch = True
        else:
            self.qc_switch = False
        self.flagStateCount = tempj['flagState']

    def log_data_in_input_json(self, timestamp, topic, data1, data2, data3):
        try:
            topic_temp_arr = topic.split('/')
            sensor_mac_add = ''.join(topic_temp_arr[2].split(':'))
            sensor_mac_add = sensor_mac_add.upper()
            s_topic = str(topic_temp_arr[1]).lower()
            try:
                flow_arr = self.sensor_mac_map_dict[sensor_mac_add][s_topic]
            except KeyError:
                return
            except IndexError:
                return
            else:
                data_arr_temp = []
                if timestamp is not None:
                    data_arr_temp.append(timestamp)
                if data1 is not None:
                    data_arr_temp.append(data1)
                if data2 is not None:
                    data_arr_temp.append(data2)
                if data3 is not None:
                    data_arr_temp.append(data3)
            try:
                for macn in self.subAssembly_input_replicate:
                    if macn.find('machine') != -1:
                        for machine_name in self.subAssembly_input_replicate[macn][flow_arr[0]]:
                            if machine_name == flow_arr[1]:
                                self.subAssembly_input_replicate[macn][flow_arr[0]][flow_arr[1]][flow_arr[2]][flow_arr[3]][flow_arr[4]].append(data_arr_temp)

            except:
                return

        except Exception as e:
            try:
                traceback.print_exc()
                print(e)
                print(timestamp, topic, data1, data2, data3)
            finally:
                e = None
                del e

    def getfiles(self, dirpath):
        try:
            a = [s for s in os.listdir(dirpath) if os.path.isfile(os.path.join(dirpath, s))]
            a.sort(key=(lambda s: os.path.getmtime(os.path.join(dirpath, s))))
            return a
        except:
            traceback.print_exc()
            return ''

    def check_files_in_raw(self, path):
        for csv_name in self.getfiles(path):
            self.makepath(csv_name, path)

    def makepath(self, csv_name, path):
        csvpath = path + csv_name[:len(csv_name)]
        if csv_name.find('.csv') != -1:
            if os.path.exists(csvpath):
                self.read_csv(csvpath, csv_name)
                time.sleep(1)

    def read_csv(self, csvpath, csv_name):
        with open(csvpath) as (csvfile):
            print('\n[+] New CSV found at: ' + csvpath)
            print('\n[+] Refetching Machine Details ... ')
            self.refetch_machine_details()
            print('\n[+] Done')
            readCSV = csv.reader(csvfile, delimiter=',')
            self.subAssembly_input_replicate = copy.deepcopy(self.subAssembly_input)
            for row in readCSV:
                cell = []
                for i in range(0, 6):
                    try:
                        cell.append(row[i])
                    except IndexError:
                        cell.append(None)

                if cell[0] == None or cell[1] == None or cell[2] == None or cell[3] == None:
                    print('\n[+] Found malformed row: ')
                    print(cell)
                else:
                    self.log_data_in_input_json(cell[0], cell[1], cell[3], cell[4], cell[5])

        os.remove(csvpath)
        print('\n[+] Length of sAIR: ' + str(len(self.subAssembly_input_replicate)))
        if len(self.subAssembly_input_replicate) > 0:
            for macid in self.subAssembly_input_replicate:
                self.subAssembly_input_replicate[macid].update({'timestamp': str(csv_name[:csv_name.find('.')])})

        else:
            self.subAssembly_input_replicate = {}
            return
        self.batch_memory_queue.append(self.subAssembly_input_replicate)
        print('\n[+] Length of BMQ: ' + str(len(self.batch_memory_queue)))
        self.subAssembly_input_replicate = {}

    def loop_memory_queue(self):
        while True:
            self.check_files_in_raw(self.bridge_data_path)
            time.sleep(2)

    def initialise_input_json(self):
        self.machine_json_dict = {}
        self.input_json_dict = {}
        self.sensor_mac_map_dict = {}
        self.machine_filepath = '/opt/novatec/data/slave/json_data/'
        machine_details_file_list_arr = glob.glob(self.machine_filepath + 'machine_details_*.json')
        for machine_details in machine_details_file_list_arr:
            config = json.loads(open(machine_details).read())
            self.machine_json_dict[machine_details] = config

        try:
            for i in self.machine_json_dict:
                temp_input_json = {}
                temp_machine_json = self.machine_json_dict[i]
                temp_input_json['header'] = {}
                if 'pumpNumber' in temp_machine_json.keys():
                    temp_input_json['header']['pumpNumber'] = int(temp_machine_json['pumpNumber'])
                temp_input_json['header']['locationName'] = str(temp_machine_json['zone']['location']['name'])
                temp_input_json['header']['companyName'] = str(temp_machine_json['zone']['location']['company']['name'])
                temp_input_json['header']['assemblyId'] = str(temp_machine_json['models']['assemblyId'])
                temp_input_json['header']['machineName'] = str(temp_machine_json['name'])
                temp_input_json['header']['manufacturer'] = str(temp_machine_json['category'])
                temp_input_json['header']['modelName'] = str(temp_machine_json['models']['name'])
                temp_input_json['header']['companyId'] = str(temp_machine_json['models']['companyId'])
                temp_input_json['header']['machineId'] = str(temp_machine_json['id'])
                temp_input_json['header']['assemblyName'] = str(temp_machine_json['models']['assembly']['name'])
                temp_input_json['header']['zoneId'] = str(temp_machine_json['zoneId'])
                temp_input_json['header']['gaugeDuration'] = int(2)
                temp_input_json['header']['zoneName'] = str(temp_machine_json['zone']['name'])
                if self.qc_switch:
                    temp_input_json['header']['group'] = 'qc'
                    try:
                        temp_input_json['header']['qc_profile'] = temp_machine_json['qcprofile']
                    except:
                        traceback.print_exc()
                        temp_input_json['header']['qc_profile'] = {}

                else:
                    temp_input_json['header']['group'] = str(temp_machine_json['group'])
                    try:
                        temp_input_json['header']['qc_profile'] = json.dumps(None)
                    except:
                        traceback.print_exc()
                        temp_input_json['header']['qc_profile'] = {}

                    for j in temp_machine_json['models']['assembly']['subAssemblyInstances']:
                        if 'subassemblies' not in temp_input_json:
                            temp_input_json['subassemblies'] = {}
                            temp_input_json['subassemblies'][str(j['name'])] = {}
                            temp_input_json['subassemblies'][j['name']]['collectors'] = {}
                        else:
                            temp_input_json['subassemblies'][str(j['name'])] = {}
                            temp_input_json['subassemblies'][j['name']]['collectors'] = {}
                    for k in temp_machine_json['sensorAssignments']:
                        temp_input_json['subassemblies'][str(k['subAssemblyInstance']['name'])]['collectors'][str(k['collectors']['name'])] = {}
                        temp_input_json['subassemblies'][str(k['subAssemblyInstance']['name'])]['collectors'][str(k['collectors']['name'])]['CSV'] = []
                        sensor_mac_temp = str(k['sensorId'].split('_')[0]).upper()
                        collector_base_name = str(k['sensorId'].split('_')[1]).lower()
                        device_name = str(k['subAssemblyInstance']['name'])
                        collector_name_temp = str(k['collectors']['name'])
                        map_array = []
                        map_array.append('subassemblies')
                        map_array.append(device_name)
                        map_array.append('collectors')
                        map_array.append(collector_name_temp)
                        map_array.append('CSV')
                        if sensor_mac_temp not in self.sensor_mac_map_dict:
                            self.sensor_mac_map_dict[sensor_mac_temp] = {}
                            self.sensor_mac_map_dict[sensor_mac_temp][collector_base_name] = map_array
                        else:
                            self.sensor_mac_map_dict[sensor_mac_temp][collector_base_name] = map_array

                    self.input_json_dict[temp_input_json['header']['machineId']] = temp_input_json

        except Exception as e:
            print("Error on load...")
            try:
                try:
                    print(e)
                    traceback.print_exc()
                finally:
                    e = None
                    del e

            finally:
                e = None
                del e


Pumpsense_Computation().main()
# okay decompiling pump-computation.pyc
