import simplejson as json
import os
import zmq

json_paths = {
    "input": "/opt/novatec/data/slave/input_json",
    "output": "/opt/novatec/data/slave/output_json"
}

input_write_enabled = True
output_write_enabled = True

zmq_send_input_json_enabled=False
zmq_send_output_json_enabled=True

def set_input_json_write(val):
    global input_write_enabled

    input_write_enabled = val

def set_output_json_write(val):
    global output_write_enabled

    output_write_enabled = val

def set_input_json_path(path):
    global json_paths
    json_paths['input'] = path

def set_output_json_path(path):
    global json_paths
    json_paths['output'] = path



jsonsender_socket = None

def zmq_init(port='4445',zmq_send_input_json=False, zmq_send_output_json=True):
    global jsonsender_socket
    global zmq_send_input_json_enabled
    global zmq_send_output_json_enabled

    zmq_send_input_json_enabled = zmq_send_input_json
    zmq_send_output_json_enabled = zmq_send_output_json

    port = str(port)
    context_jsonsender = zmq.Context()
    jsonsender_socket = context_jsonsender.socket(zmq.PUSH)
    jsonsender_socket.connect("tcp://localhost:" + port)

def zmq_send_json(filename,data,file_type="output"):
    global jsonsender_socket

    if not jsonsender_socket:
        return
    socket_string=f"openwrt~~{file_type}~~{filename}~~{json.dumps(data,ignore_nan=True)}"
    jsonsender_socket.send_string(socket_string)


def json_file_store(path,file_name,json_parsed):
  
    if not os.path.exists(path):
        os.mkedirs(path, exists_ok=True)
    json_file=os.path.join(path,file_name)

    with open(json_file,"w") as f:
        json.dump(json_parsed,f,indent=4,ignore_nan=True)


def store_input_json(json_parsed, force=False):
    global input_json_write
    global json_paths

    file_name=f"{json_parsed['header']['timestamp']}_{json_parsed['header']['machineId']}_input.json"
    if input_write_enabled or force:
        json_file_store(json_paths['input'],file_name,json_parsed)
    if zmq_send_input_json_enabled:
        zmq_send_json(file_name,json_parsed,file_type="input")

def store_output_json(json_parsed, force=False):
    global output_json_write
    global json_paths

    #file_name=f"{json_parsed['header']['timestamp']}_machine_qc_{json_parsed['header']['machineId']}_output.json"
    file_name=f"{json_parsed['header']['timestamp']}_{json_parsed['header']['machineId']}_output.json"
    if output_write_enabled or force:
        json_file_store(json_paths['output'],file_name,json_parsed)
    if zmq_send_output_json_enabled:
        zmq_send_json(file_name,json_parsed,file_type="output")

