###################################################################################################
# File name: pump_analytics.py
# Created on: July 17, 2020
# Copyright: MachineSense
# Author: Sai Ram, Biswadeep Ghoshal
# Description: This is a python3 script, this script creates output JSON. It takes 
#              input JSON and creates the output JSON.        
###################################################################################################

import time, glob
import platform
from platform import system
from copy import deepcopy
import json
import sys
import os
import traceback
from pyanalytics_lib import JsonDataStore
from pygaugelib.JsonDetailConfigs import fetch_gauge_rule_list

import warnings
warnings.filterwarnings("ignore")

from qc_computation.gauge_code.Input_Processing import AC_Feature_Extractions_From_Input, XL_Feature_Extractions_From_Input, VAC_Feature_Extractions_From_Input
from qc_computation.gauge_code.Gauge_Processing_qc import *
from qc_computation.gauge_code.vmart_output import *
from qc_computation.gauge_code.stored_objects_qc import * #stored_object, Stored_Object_Saving
#from qc_computation.gauge_code.log_info_updation import read_log_info, add_log_info, remove_old_log_info, write_log_info

from pygaugelib.json_logger import json_logger_init, add_log_info, json_logger_save , json_logger_purge
from pygaugelib.api_processing_2 import *
from pygaugelib.additional_functions_ver_1_12 import *
from pygaugelib.exception_handler import *
from pygaugelib.Outlier import Outlier_Indicator

#from DatahubQCLogger import AnalyticsLogger

#logger = AnalyticsLogger('localhost:5000', '/usr/src/conf/qc_input_configuration.json')   

logger = print

##################################################################################################

def qc_init(env):
    global json_details_path
    global working_dir
    global log_file
    global config_path
    global config_data
    global stored_objects_dir
    global conf_dir

    config_data={}
    json_details_path = env.get('json_details_path','/opt/novatec/slave/json_data')
    working_dir = env.get('working_dir','/tmp/qc_computation/')
    log_file=env.get('log_file','/tmp/qc/log/reason_log.json')
    stored_objects_dir=env.get('stored_objects_dir','/tmp/qc/stored_objects')
    conf_dir = env.get('conf_dir',"/tmp/qc/conf")
    config_path=env.get('config_path',"/tmp/qc/conf/r_config.json")
    config_data['write_qc_input_json'] = env.get('write_qc_input_json',False)
    config_data['write_qc_output_json'] = env.get('write_qc_output_json',False)
    config_data['json_sender_port'] = env.get('json_sender_port',"4445")
    config_data['zmq_send_enabled'] = env.get('zmq_send_enabled',False)
    config_data['zmq_send_input_json'] = env.get('zmq_send_input_json',False)
    config_data['zmq_send_output_json'] = env.get('zmq_send_output_json',True)

    input_json_path = env.get("INPUT_JSON_FOLDER_PATH",os.path.join(working_dir,"input_json"))
    output_json_path = env.get("OUTPUT_JSON_FOLDER_PATH",os.path.join(working_dir,"output_json"))
    JsonDataStore.set_input_json_path(input_json_path)
    JsonDataStore.set_input_json_write(config_data['write_qc_input_json'])
    JsonDataStore.set_output_json_path(output_json_path)
    JsonDataStore.set_output_json_write(config_data['write_qc_output_json'])

    qc_startup()

    
####################################################################################################
##                           Config Json File Existence and Reading Data
####################################################################################################

    

def qc_startup():

    os.chdir(working_dir)
    json_logger_init(log_file)        #.. To store log of errors or issues ..#

    #hy_data_rate_per_sec = config_info['hy_data_rate_per_sec']
    #ac_data_rate_per_sec = config_info['ac_data_rate_per_sec']
    #ir_data_rate_per_sec = config_info['ir_data_rate_per_sec']


    if config_data.get('zmq_send_enabled'):
        JsonDataStore.zmq_init(port=config_data.get('json_sender_port',"4445"),zmq_send_input_json=config_data.get('json_send_input_json'), zmq_send_output_json=config_data.get('json_send_output_json'))
  

###################################################################################################
#                           Data Processing After Receiving Json Files
###################################################################################################


def main_output_qc(json_parsed):

    print("Called QC Main function")
    current_datetime_stamp=int(time.time())
    version_product=None
    input_json_write=config_data.get("write_qc_input_json",False)
    output_json_write=config_data.get("write_qc_output_json",False)
    end_qc=False
    
    try:
        
        if len(json_parsed) != 0:

            try:
               JsonDataStore.store_input_json(json_parsed)
            except Exception as e:
                print("Error Writing Input JSON: ",str(e))
            #... Input json header processing ...#
            #json_parsed["header"]["qc_profile"]["flagStateCount"]=0
            current_datetime_stamp = int(json_parsed['header']['timestamp'])
            group = json_parsed['header']['group']                   
            manufacturer = json_parsed['header']['manufacturer'].lower()
            modelName = json_parsed['header']['modelName']
            model="VPDB" if "VPDB" in modelName else "SVP" if "SVP" in modelName else "MVP" if "MVP" in modelName else "Default"
            subassembly_instance_list = list(json_parsed['subassemblies'].keys())
            assembly = json_parsed['header']['assemblyName'].lower()

            version_product="301.0000" if any(["componentanalyzer" in x.lower() for x in subassembly_instance_list]) else "302.0000"
            
            
            #... stored_objects_initialization ...#

            machineid = json_parsed["header"]["machineId"]

            #stored_object_path="F:/MachineSense/qc_computation/stored_objects/stored_objects_{}.json".format(machineid) if system()=="Windows" else "/tmp/qc/stored_objects/stored_objects_{}.json".format(machineid)
            stored_object_path=os.path.join(stored_objects_dir,f"stored_objects_{machineid}.json")
            #stored_object_path=os.path.join("/tmp/qc/stored_objects",f"stored_objects_{machineid}.json")

            stored_object=read_stored_object(stored_object_path)

            
            #... Fetching the gauge details from the gauge details json files in the datahub ...#
       
            gauge_rule_list = fetch_gauge_rule_list(machineid,json_details_path=json_details_path)  
                                                                 
            count_qc=0                                                      #.. This count is defined just to know whether required collector data are present in Input JSON ..#
            #... Starting to Process Input Data to Publish Gauges and VMart Stats ...#
            
            if len(subassembly_instance_list) != 0:
                i = 0; combined_gauge_output = {}; combined_vmart_output = {}
                for i in range(len(subassembly_instance_list)):
                    subassembly_instance = subassembly_instance_list[i]
                    try:
                        subassembly = json_parsed['subassemblies'][subassembly_instance]['header']['subAssembly']['name'].lower()
                    except:
                        subassembly="".join([i for i in subassembly_instance if i.isalpha()]).lower()

                    if json_parsed['subassemblies'][subassembly_instance]['collectors']!={}:
                        collector_names = list(json_parsed['subassemblies'][subassembly_instance]['collectors'].keys())
                        
                        ac_key_indicator = int( 'ac1' in collector_names )
                        ir_key_indicator = int( 'ir1' in collector_names )
                        hy_key_indicator = int( 'hy1' in collector_names )
                        xl_key_indicator = int( 'xl1' in collector_names )
                        vac_key_indicator = int( 'vac1' in collector_names )
                        bp_key_indicator = int( 'bp1' in collector_names )
                        
                        
                        if group.lower() == "qc" and json_parsed["header"].get("qc_profile",{}) not in [{},None,"null"]:
                            ir_stat_list=None
                            hy_stat_list=None
                            bp_stat_list=None
                            
                            
                            if ac_key_indicator == 1 and isinstance(json_parsed['subassemblies'][subassembly_instance]['collectors']['ac1'],dict) and "CSV" in json_parsed['subassemblies'][subassembly_instance]['collectors']['ac1'] and json_parsed['subassemblies'][subassembly_instance]['collectors']['ac1']['CSV']!=[]:
                                ac_stat_list = AC_Feature_Extractions_From_Input(json_parsed,subassembly_instance,gauge_rule_list,current_datetime_stamp,stored_object)
                                outlier_indicator_ac = Outlier_Indicator(ac_stat_list) if ac_stat_list is not None else None
                                machine_stationary_ac = Machine_Stationary( ac_stat_list ) if ac_stat_list is not None else None
                               
                            else:
                                ac_stat_list = None; machine_stationary_ac = None; outlier_indicator_ac = None; count_qc+=1
                                
                            if xl_key_indicator == 1 and isinstance(json_parsed['subassemblies'][subassembly_instance]['collectors']['xl1'],dict) and "CSV" in json_parsed['subassemblies'][subassembly_instance]['collectors']['xl1'] and json_parsed['subassemblies'][subassembly_instance]['collectors']['xl1']['CSV']!=[]:
                                xl_stat_list = XL_Feature_Extractions_From_Input(json_parsed,subassembly_instance,gauge_rule_list,current_datetime_stamp,stored_object)
                                outlier_indicator_xl = Outlier_Indicator(xl_stat_list) if xl_stat_list is not None else None
                                machine_stationary_xl = Machine_Stationary( xl_stat_list ) if xl_stat_list is not None else None

                            else:
                                xl_stat_list = None; machine_stationary_xl = None; outlier_indicator_xl = None; count_qc+=1
                                
                            if vac_key_indicator == 1 and isinstance(json_parsed['subassemblies'][subassembly_instance]['collectors']['vac1'],dict) and "CSV" in json_parsed['subassemblies'][subassembly_instance]['collectors']['vac1'] and json_parsed['subassemblies'][subassembly_instance]['collectors']['vac1']['CSV']!=[]:
                                vac_stat_list = VAC_Feature_Extractions_From_Input( json_parsed,subassembly_instance,gauge_rule_list,stored_object,current_datetime_stamp)
                                machine_stationary_vac = Machine_Stationary_VAC( vac_stat_list,subassembly_instance,stored_object,gauge_rule_list,current_datetime_stamp ) if vac_stat_list is not None else None
                            else:
                                vac_stat_list = None
                                machine_stationary_vac = None
                                count_qc+=1
                            print("Calling Gauge_Computation from main_output_qc: ", model,assembly,subassembly,subassembly_instance)    
                            gauge_output = Gauge_Computation(json_parsed,model,assembly,subassembly,subassembly_instance,manufacturer,gauge_rule_list,stored_object,ac_stat_list,xl_stat_list,ir_stat_list,hy_stat_list,vac_stat_list,current_datetime_stamp,outlier_indicator_ac,machine_stationary_ac,outlier_indicator_xl,machine_stationary_xl,machine_stationary_vac)
                            vmart_output = Vmart(json_parsed, subassembly_instance, ac_stat_list, ac_key_indicator, xl_stat_list, xl_key_indicator,vac_stat_list, vac_key_indicator, ir_stat_list, ir_key_indicator, hy_stat_list, hy_key_indicator, bp_stat_list, bp_key_indicator)

                            
                            
# =============================================================================
#                             elif group.lower() != "qc":
#                                 gauge_output={}
#                                 vmart_output={}
#                                 add_log_info(current_datetime_stamp,"main","normal","Unidentified group '{}'".format(group))
#                                 
# =============================================================================
                        else:
                            gauge_output={}
                            vmart_output={}
                            add_log_info(current_datetime_stamp,"main","normal","QC Profile Missing or Some Other Group {}".format(group))

                        
                        combined_gauge_output.update(gauge_output)
                        combined_vmart_output.update(vmart_output)
                        
                #... Gauge and VMart Stats Published ...#

                #... Output JSON ...#
                json_parsed["header"]["timestamp"]=int(json_parsed["header"]["timestamp"])            
                final_output={'header':json_parsed["header"],
                              'gauge': combined_gauge_output,
                              'machine_parameters' : None,
                              'subassemblies': combined_vmart_output,
                              'version':version_product,
                              'last modified':"13 July 2021"}

                #... Writing Stored Object For Use in Processing Further Batch of Data ...#
                print("Stored object path: ",stored_object_path)
                Stored_Object_Saving(stored_object,stored_object_path)
                print("stored object updated successfully")
                
                                
                #... Creating Socket String For Output JSON ...#
                filename=str(json_parsed["header"]["timestamp"])+"_"+json_parsed["header"]["machineId"]+".json"
                socket_string="R~~output~~{}~~{}".format(filename,json.dumps(final_output))
                #print("Final Output: ",json.dumps(final_output,indent=4))
                #... Sending Output JSON ...#
                if any([json_parsed['subassemblies'][subassembly_instance]['collectors']!={} for subassembly_instance in subassembly_instance_list]) and count_qc!=3*len(subassembly_instance_list):
                    try:
                        JsonDataStore.store_output_json(final_output)
                    except Exception as e:
                        print("Error Writing Output JSON: ",str(e))
                    #jsonsender_socket.send_string(socket_string)
                    print("Output JSON sent successfully")
                else:
                    add_log_info(current_datetime_stamp,"main_output","normal","{}: Reqd Collector Data Not Present in Input JSON".format(json_parsed["header"]["machineId"]))
                    
            else:
                final_output = None
                add_log_info(current_datetime_stamp,"main_output","normal","{}: No Subassembly in Input JSON".format(json_parsed["header"]["machineId"]))
                
            
                
            if stored_object['vacuum_qc']['qc_gauge_progress_percent']==100 and stored_object['vibration_qc']["qc_gauge_progress_percent"]==100:
                
               
                if stored_object['vacuum_qc']['qc_gauge_green_percent'] >= 30 and stored_object['vibration_qc']['qc_gauge_green_percent'] >= 30:
                    #logger.log('Overall Machine QC process completed', 'success', True, True)
                    print('Overall Machine QC process completed', 'success', True, True)
                    
                
                else:
                    print('Overall Machine QC process completed', 'success', True, False)
                
                
                end_qc=True
   
            
            #.. Else, if Vacuum QC Process Is Complete But Vibration QC has not started yet, meaning the AC Sensor is disconnected ..#
            elif stored_object['vacuum_qc']['qc_gauge_progress_percent']==100 and stored_object['vibration_qc']['qc_gauge_progress_percent']==0:
                
                if stored_object['vacuum_qc']['qc_gauge_green_percent'] >= 30:
                    #logger.log('Overall Machine QC process completed only with Vacuum', 'success', True, True)
                    print('Overall Machine QC process completed only with Vacuum', 'success', True, True)
                
                else:
                    #logger.log('Overall Machine QC process completed only with Vacuum', 'success', True, False)
                    print('Overall Machine QC process completed only with Vacuum', 'success', True, False)
                
                end_qc=True
                    
            
            #.. Else, if Vibration QC Process Is Complete But Vacuum QC has not started yet, meaning the VAC Sensor is disconnected ..#
            elif stored_object['vacuum_qc']['qc_gauge_progress_percent']==0 and stored_object['vibration_qc']['qc_gauge_progress_percent']==100:
                
                if stored_object['vibration_qc']['qc_gauge_green_percent'] >= 30:
                    #logger.log('Overall Machine QC process completed only with Vibration', 'success', True, True)
                    print('Overall Machine QC process completed only with Vibration', 'success', True, True)
                
                else:
                    #logger.log('Overall Machine QC process completed only with Vibration', 'success', True, False)
                    print('Overall Machine QC process completed only with Vibration', 'success', True, False)
                
                end_qc=True
                    
            
            #.. If none of the above are satisfied, it means QC is still in progress
            else:    
                #logger.log('Overall Machine QC process completion pending', 'success', False, None)       
                print('Overall Machine QC process completion pending', 'success', False, None)       
        else:
            final_output = None
            add_log_info(current_datetime_stamp,"main_output","normal","Input JSON is blank")


    except Exception:
        error_text=traceback.format_exc()
        add_log_info(current_datetime_stamp,"main_output","error",error_text)
        traceback.print_exc()

        #... Writing Input JSON and Output JSON for debugging purposes ...#
        
        try: 
            JsonDataStore.store_input_json(json_parsed)
        except Exception as e:
            print("Error Writing JSON: ",str(e))
                
    if version_product=="302.0000":
        version_file = os.path.join(conf_dir,"analytics_version_vpa.txt")
        with open(version_file, "w") as v_file:
            v_file.write("version={}".format(version_product))

    else:
        version_file = os.path.join(conf_dir,"analytics_version_ca.txt")
        with open(version_file, "w") as v_file:
            v_file.write("version={}".format(version_product))

        
    json_logger_purge(current_datetime_stamp)        #.. Remove issues or errors older than 24 hours from the log ..#
    json_logger_save()
    
    #... Stopping QC Service When Machine QC completes ...#
    if end_qc:
        #logger.update_qc_input_configuration(False)
        print("end_qc update_qc_input_configuration: ",False)
    return final_output 

