

''' This module contains functions which call and compute gauges to be published in Output JSON '''


import traceback
from copy import deepcopy
from pygaugelib.exception_handler import Numeric_Input_Type_Checking, Non_Negative_Out_Of_Range_Exception_Checking


global stored_object 
from numpy import array, nanmedian, nanstd, greater
from scipy.signal import argrelextrema
#from DatahubQCLogger import AnalyticsLogger


#from qc_computation.gauge_code.log_info_updation import add_log_info

from pygaugelib.kernel_density_user_defined_3 import kde
from pygaugelib.additional_functions_ver_1_12 import Machine_Stationary
from pygaugelib.json_logger import add_log_info


#logger = AnalyticsLogger('localhost:5000', '/usr/src/conf/qc_input_configuration.json')   
logger = print

def Vibration_Quality_Control(ac_stat_list,subassembly_instance,subassembly,current_datetime_stamp):
    try:
        raw_data = ac_stat_list['raw_data']      #...... Data frame containing r, theta and phi
        if len(raw_data[:,0])>2:
            r_density = kde( raw_data[:,0] )
            #r_density=kde(raw_data)
            arg=r_density['arguments']
            dens=array(r_density['densities'])
            median=nanmedian(arg)
            stdev=nanstd(arg,ddof=1)
            index_centre=[index for index,value in enumerate(arg) if value>median-stdev and value<median+stdev]
            area_at_centre=sum(dens[index_centre])
            index_tail=[index for index,value in enumerate(arg) if value<=median-(1.5*stdev) or value>=median+(1.5*stdev)]
            area_at_tail=sum(dens[index_tail])
            
            rtc_10=10*(area_at_tail/area_at_centre)
            rtc_10_color='green' if rtc_10<=0.075 else 'red'
            #===check this========
            #index=max(argrelextrema(dens,np.greater)[0])
            local_max=argrelextrema(dens,greater)[0]
            index=max(local_max)
            
            area_below_max_peak=sum(dens[:index])/sum(dens)
            area_above_max_peak=sum(dens[(index+1):len(dens)])/sum(dens)
            abs_area_of_asymmetry=abs(area_above_max_peak-area_below_max_peak)
            abs_area_of_asymmetry_color='green' if abs_area_of_asymmetry<=0.6 else 'red'
            vib_qc_color='red' if rtc_10_color=='red' and abs_area_of_asymmetry_color=='red' else 'green'
            combined_input_exception_indicator = 0
        else:
            combined_input_exception_indicator = 1
        machine_stationary_status = Machine_Stationary( ac_stat_list )
        
        if machine_stationary_status == "Machine Idle":
            output= {"color":["red"], "value":[-1],"stats":{'rtc_10':rtc_10,'abs_area_of_asymmetry':abs_area_of_asymmetry}, "trendExistence":False,"purpose":"qc","success":"true","reason":"MachineIdle" }
        elif combined_input_exception_indicator == 0:
            output= {"color":[vib_qc_color], "value":[-1],"stats":{'rtc_10':rtc_10,'abs_area_of_asymmetry':abs_area_of_asymmetry}, "trendExistence":False, "purpose":"qc","success":"true" }
        else:
            output= {"color": ["red"],"value":[-1],"trendExistence":False,"purpose":"qc", "success":"true","reason":"Insufficient Data"}
            
    except Exception:
        error_text=traceback.format_exc()
        add_log_info(current_datetime_stamp,"Vibration_Quality_Control","error",error_text)
        output= {"color": ["green"],"value":[-1],"trendExistence":False,"purpose":"qc", "success":"false","reason":"Insufficient Data"}
    
    output.update({'subassemblyInstance':subassembly_instance, 'subassembly':subassembly})
    
    return output
    
    
    
def Vacuum_Quality_Control(vac_stat_list,subassembly_instance,subassembly,gauge_rule_list=None,current_datetime_stamp=0):                     #intermediate_stored_objects, , subassembly_instance, gauge_rule_list):
    try:
        
        if gauge_rule_list is None:
            gauge_rule_list={}
            
        vac_load_cutoff = int(gauge_rule_list.get("Vacuum Pump Utilization Under Load (24 hrs)",{}).get(subassembly_instance,{}).get("load_threshold",2))
            
        if Numeric_Input_Type_Checking( [vac_stat_list['high_vac']] ) == 0:
            vac_machine_stationary_indicator = 'No' if  vac_stat_list['high_vac'] > vac_load_cutoff else 'Yes'
        else:
            vac_machine_stationary_indicator = 'No'            
            
        
        #if not (vac_machine_stationary_indicator == 'Yes' or vac_stat_list['vac_data_status'] == 'Insufficient Data'):
        batch_high_vac=vac_stat_list["high_vac"] 
        
        #........... One Time specification needed ........
        non_autobaseline_numeric_inputs = [batch_high_vac]
        
        #....... Exception Handling .............. 0 for no errors and 1 for at least one error ........
        non_autobasline_numeric_inputs_exception_indicator = Numeric_Input_Type_Checking(non_autobaseline_numeric_inputs)
        numeric_out_of_range_exception_indicator = Non_Negative_Out_Of_Range_Exception_Checking(non_autobaseline_numeric_inputs )
        
        combined_input_exception_indicator=0 if numeric_out_of_range_exception_indicator+non_autobasline_numeric_inputs_exception_indicator==0 else 1
        
        if combined_input_exception_indicator==0:
            
            #========================need to discuss==============================
            
            vacuum_qc_color='green' if (batch_high_vac>=5 and batch_high_vac<=7) or (batch_high_vac>=9 and batch_high_vac<=13) else 'red'
            
        #===========to check for idle condition of the machine================       
        #machine_stationary_vac_called = MachineStationary_status_vac( input_json, intermediate_stored_objects, vac_stat_list, subassembly_instance, gauge_rule_list)
        #machine_stationary = machine_stationary_vac_called['vac_machine_stationary_indicator']
        output_conditions={'combined_input_exception_indicator':combined_input_exception_indicator}#, 'machine stationary':machine_stationary}

        if output_conditions['combined_input_exception_indicator']==1:
            output={"color":["red"], "value":[-1], "trendExistence":False, "success":"true", "purpose":"qc", "reason":"High Vac calculation not possible due to insufficient data"}
        #elif output_conditons['machine stationary']=='Yes':
            #output={"color":"red", "value":[-1], "stats":{"high_vac" : batch_high_vac }, "trendExistence":False, "success":"true", "purpose":"qc", "reason":"Low Vac Detected"}
        elif vac_machine_stationary_indicator == 'Yes':
            output={"color":["red"], "value":[-1], "stats":{"high_vac":batch_high_vac}, "trendExistence":False, "success" : "true", "purpose":"qc" ,"reason" :"Low Vac Detected" }
        else:
            output= {"color":[vacuum_qc_color], "value":[-1], "stats":{"high_vac":batch_high_vac}, "trendExistence":False, "success" : "true", "purpose":"qc"}
        
        
        
#=================need to write catch========================
        
    except Exception:
        error_text=traceback.format_exc()
        add_log_info(current_datetime_stamp,"Vacuum_Quality_Control","error",error_text)
        output = {"color": ["green"],"value":[-1],"purpose":"qc", "success":"false", "trendExistence": False,"reason" : 'Insufficient Data.'}
 
    output.update({'subassemblyInstance':subassembly_instance, 'subassembly':subassembly})
    
    return output
 
def QC_Gauge_Output( subassembly_instance,Vibration_QC_output,Vacuum_QC_output,ac_stat_list,vac_stat_list,current_datetime_stamp):
    try:
        
        final_output = {}
        
        if ac_stat_list is not None:
            gauge_name_list = subassembly_instance + ":  " + 'Vibration QC'
            final_output[gauge_name_list] =  deepcopy(Vibration_QC_output)  
    
        if vac_stat_list is not None:
            gauge_name_list = subassembly_instance +':  ' + 'Vacuum QC'
            final_output[gauge_name_list] = deepcopy(Vacuum_QC_output)
            

        return final_output
    except Exception:
        error_text=traceback.format_exc()
        add_log_info(current_datetime_stamp,"QC_Gauge_Output","error",error_text)
        return {}
    
    
def QC_Output_Processing(input_json,stored_object,subassembly_instance,qc_gauge_output):
    qc_gauge_name="vibration_qc" if "machinesense" in subassembly_instance.lower() else "vacuum_qc"
    qc_external_gauge_name=subassembly_instance + ":  " + ('Vibration QC' if "machinesense" in subassembly_instance.lower() else 'Vacuum QC')
    if 'previous_state_count' not in stored_object[qc_gauge_name]:
        stored_object[qc_gauge_name]["previous_state_count"] = 0

        
    current_qc_state_count=int(input_json["header"].get("qc_profile",{}).get("flagStateCount",0))
    
    previous_qc_state_count = stored_object[qc_gauge_name]['previous_state_count']
    
    if current_qc_state_count != previous_qc_state_count:   #..... if state_count increases => change in state has occurred => reset gauges
        reset_gauge_indicator = "Yes"
        
# =============================================================================
#         if current_qc_state_count%2==0 and previous_qc_state_count%2==1:
#             if "qc_complete" not in stored_object or stored_object["qc_complete"]:
#                 stored_object["qc_complete"]=False
# =============================================================================
    else:
        reset_gauge_indicator = "No"    #.... if no change in state_count => state is same => gauges may or may not reset. For now initialize to "No"
    
        
    
    if  current_qc_state_count % 2 == 1:               #..... if odd reset gauge and don't progress i.e. NULL output
        reset_gauge_indicator = "Yes"  
        gauge_null_output = "Yes"     #..... Note: previous reset_gauge_indicator value is overwritten and it should be
    else:
        gauge_null_output = "No"                                #.... if even there is non-NULL output
  
    if reset_gauge_indicator == "Yes":
        
        if qc_external_gauge_name in qc_gauge_output:
        
            qc_gauge_output[qc_external_gauge_name]["progress"] = 0          #..... progress is reset to 0 in gauge_output
        
        
        stored_object[qc_gauge_name]['qc_gauge_progress_percent'] = 0    #....... progress is reset to 0 in stored_object
    
    
        stored_object[qc_gauge_name]['qc_gauge_green_percent'] = 0     #....... green percent is reset to 0 in stored_object
    
    
        stored_object[qc_gauge_name]['qc_gauge_red_percent'] = 0     #....... red percent is reset to 0 in stored_object
    
   #...... Updating previous state_count in stored_object with current values ........
  
    stored_object[qc_gauge_name]["previous_state_count"] = current_qc_state_count
  
  
  #...... Final processing ......
  
    if  gauge_null_output == "No":    #.... case when there is no NULL output
    
        return { 'qc_gauge_output':qc_gauge_output, 'updated_stored_objects':stored_object }
    
    else:         #....... case when there is NULL output
    
        return  {'qc_gauge_output': {}, 'updated_stored_objects': stored_object }
    

#... Mapping Internal Gauge Names To External Gauge Names For Color Modification With Respect To Gauge User Inputs ...#


def update_qc_gauge_stored_object(QC_gauge_output, so):

    QC_gauge_progress_percent = so['qc_gauge_progress_percent']
    #QC_gauge_green_percent = so['qc_gauge_green_percent']
    #QC_gauge_red_percent = so['qc_gauge_red_percent']

    if QC_gauge_progress_percent<100:  
        #==================need to discuss============================
        if QC_gauge_output['success']=='true':
            so['qc_gauge_progress_percent'] += 10
                 
            if QC_gauge_output['color']==['green']: 
                so['qc_gauge_green_percent']+=10
                #logger.log('Vibration Machine QC process running : [Green] ' + str(stored_object['vibration_qc']['qc_gauge_green_percent'])+ " %", 'success', False, None)
                 
            if QC_gauge_output['color']==['red']:
                so['qc_gauge_red_percent']+=10
                #logger.log('Vibration Machine QC process running : [Red] ' + str(stored_object['vibration_qc']['qc_gauge_red_percent'])+ " %", 'failed', False, None)
            
        if so['qc_gauge_green_percent']>=30:
            so['qc_gauge_progress_percent']=100
            
        if so['qc_gauge_red_percent']>=80:
            so['qc_gauge_progress_percent']=100
            
        if  so['qc_gauge_progress_percent']==100:
                 
            if so['qc_gauge_red_percent'] >= 80:
                QC_gauge_output['color']=['red']
                #logger.log('Vibration Machine QC process completed => Not Ok', 'success', False, None)
            else: 
                QC_gauge_output['color']=['green']   
                #logger.log('Vibration Machine QC process completed => Ok', 'success', False, None)


def update_mvp_qc_gauge(QC_gauge_output, so,model):
    if model == "MVP":
        QC_gauge_output["progress"] = 100
        QC_guage_output["color"] = ['green']
        QC_gauge_output["reason"] = None
        so["qc_gauge_progress_percent"] = 100


def Gauge_Vibration_QC(model,ac_stat_list,xl_stat_list,stored_object,subassembly_instance,subassembly,current_datetime_stamp):
    ####### Vibration QC #######

    so = stored_object['vibration_qc']
    stat_list=ac_stat_list if ac_stat_list is not None else xl_stat_list
    if so['qc_gauge_progress_percent'] >= 100:  # We are done so we return back an empty gauge.
        #return stat_list,{}
        return None,{}	# Force stat list to be None here since that is what the flow did before, not sure why.
    if not stat_list:
        return stat_list,{}
    
                
    Vibration_QC_output = Vibration_Quality_Control(stat_list,subassembly_instance,subassembly,current_datetime_stamp)
    update_qc_gauge_stored_object(Vibration_QC_output,so)

    #=========================need to do======================================
    Vibration_QC_output['progress']=so['qc_gauge_progress_percent']
    update_mvp_qc_gauge(Vibration_QC_output, so,model) 
    return stat_list,Vibration_QC_output
            
def Gauge_Vacuum_QC(model,vac_stat_list,stored_object,subassembly_instance,subassembly,gauge_rule_list,current_datetime_stamp):
    so = stored_object['vacuum_qc']
    if so['qc_gauge_progress_percent'] >= 100:  # We are done so we return back an empty gauge.
        return {}

    if vac_stat_list is None:
        return {}
                
    #vac_stat_list = VAC_Feature_Extractions_From_Input(input_json,i)

    #===================this needs to be done===========================
    Vacuum_QC_output=Vacuum_Quality_Control(vac_stat_list,subassembly_instance,subassembly,gauge_rule_list,current_datetime_stamp)
    update_qc_gauge_stored_object(Vacuum_QC_output,so)

    Vacuum_QC_output['progress']=so['qc_gauge_progress_percent']
    update_mvp_qc_gauge(Vacuum_QC_output, so,model) 
            
    return Vacuum_QC_output


# Calculation of Gauges for Meta-Parameters
def Gauge_Computation(input_json,model,assembly,subassembly,subassembly_instance,manufacturer,gauge_rule_list,stored_object,ac_stat_list,xl_stat_list,ir_stat_list,hy_stat_list,vac_stat_list,current_datetime_stamp,outlier_indicator_ac,machine_stationary_ac,outlier_indicator_xl,machine_stationary_xl,machine_stationary_vac):
    print("Gauge_Computation Starting...")
    try:
        
        manufacturer_list = ['oem','retrofit']
        #list_of_subassembly = ['blower','componentanalyzer','machinesense','cdablower','processblower','genericmotor','advancedcomponentanalyzer','regenblower','machinebody','vfd','cassettemotor','gearbox','nvfd','motorpanel','boosterblower','vacuumsense','vacuumanalyzer']
        #assembly_list = ['All','componentanalyzer','vacuum-pump','advancedcomponentanalyzer','machinesense','railcar','vacuumanalyzer']
        
        #machineid = input_json["header"]["machineId"]

        
        if manufacturer in manufacturer_list:
                        
                
            ####### Vibration QC #######
            stat_list,Vibration_QC_output = Gauge_Vibration_QC(model,ac_stat_list,xl_stat_list,stored_object,subassembly_instance,subassembly,current_datetime_stamp)
            ####### Vacuum QC #######
            Vacuum_QC_output = Gauge_Vacuum_QC(model,vac_stat_list,stored_object,subassembly_instance,subassembly,gauge_rule_list,current_datetime_stamp)
########   End VACUUM QC ****************           

            print("Vibration_QC_Output: ",Vibration_QC_output)
            print("Vacuum QC Output: ", Vacuum_QC_output) 
            qc_gauge_output = QC_Gauge_Output(subassembly_instance,Vibration_QC_output,Vacuum_QC_output,stat_list,vac_stat_list,current_datetime_stamp)
            processed_gauge_output = QC_Output_Processing( input_json, stored_object, subassembly_instance, qc_gauge_output )
            gauge_output=processed_gauge_output['qc_gauge_output']
            
            
            #... Individual QC Process: Vibration ...#
            
            if 0 < stored_object['vibration_qc']['qc_gauge_progress_percent'] < 100:
                
                if Vibration_QC_output != {}:
                    
                    try:
                        if Vibration_QC_output["color"]==["green"]:
                            #logger.log('Vibration Machine QC process running : [Green] ' + str(stored_object['vibration_qc']['qc_gauge_green_percent'])+ " %", 'success', False, None)
                            print('Vibration Machine QC process running : [Green] ' + str(stored_object['vibration_qc']['qc_gauge_green_percent'])+ " %", 'success', False, None)
                        elif Vibration_QC_output["color"]==["red"]:
                            #logger.log('Vibration Machine QC process running : [Red] ' + str(stored_object['vibration_qc']['qc_gauge_red_percent'])+ " %", 'failed', False, None)
                            print('Vibration Machine QC process running : [Red] ' + str(stored_object['vibration_qc']['qc_gauge_red_percent'])+ " %", 'failed', False, None)
                            
                    except Exception:
                        error_text=traceback.format_exc()
                        add_log_info(current_datetime_stamp,"Gauge_Computation","error",error_text+"\n\nVibration_QC_output: {}".format(Vibration_QC_output))
                        
            elif stored_object['vibration_qc']['qc_gauge_progress_percent'] == 100:
                if Vibration_QC_output != {}:
                    try:
                        if Vibration_QC_output["color"]==["green"]:
                            #logger.log('Vibration Machine QC process completed => Ok', 'success', False, None)
                            print('Vibration Machine QC process completed => Ok', 'success', False, None)
                        elif Vibration_QC_output["color"]==["red"]:
                            #logger.log('Vibration Machine QC process completed => Not Ok', 'success', False, None)
                            print('Vibration Machine QC process completed => Not Ok', 'success', False, None)
                    except Exception:
                        error_text=traceback.format_exc()
                        add_log_info(current_datetime_stamp,"Gauge_Computation","error",error_text+"\n\nVibration_QC_output: {}".format(Vibration_QC_output))
            
# =============================================================================
#             else:
#                 logger.log('Vibration Machine QC process initialized', 'success', False, None)
# =============================================================================

            #... Individual QC Process: Vacuum ...#  
            
            if 0 < stored_object['vacuum_qc']['qc_gauge_progress_percent'] < 100:
                
                if Vacuum_QC_output != {}:
                    try:
                        if Vacuum_QC_output["color"]==["green"]:
                            #logger.log('Vacuum Machine QC process running : [Green] ' + str(stored_object['vacuum_qc']['qc_gauge_green_percent'])+ " %", 'success', False, None)
                            print('Vacuum Machine QC process running : [Green] ' + str(stored_object['vacuum_qc']['qc_gauge_green_percent'])+ " %", 'success', False, None)
                        elif Vacuum_QC_output["color"]==["red"]:
                            #logger.log('Vacuum Machine QC process running : [Red] ' + str(stored_object['vacuum_qc']['qc_gauge_red_percent'])+ " %", 'failed', False, None)
                            print('Vacuum Machine QC process running : [Red] ' + str(stored_object['vacuum_qc']['qc_gauge_red_percent'])+ " %", 'failed', False, None)
                            
                    except Exception:
                        error_text=traceback.format_exc()
                        add_log_info(current_datetime_stamp,"Gauge_Computation","error",error_text+"\n\nVacuum_QC_output: {}".format(Vacuum_QC_output))

            
            elif stored_object['vacuum_qc']['qc_gauge_progress_percent'] == 100:
                
                if Vacuum_QC_output != {}:
                    try:
                        if Vacuum_QC_output["color"]==["green"]:
                            #logger.log('Vacuum Machine QC process completed => Ok', 'success', False, None)
                            print('Vacuum Machine QC process completed => Ok', 'success', False, None)
                        elif Vacuum_QC_output["color"]==["red"]:
                            #logger.log('Vacuum Machine QC process completed => Not Ok', 'success', False , None)
                            print('Vacuum Machine QC process completed => Not Ok', 'success', False , None)
                    except Exception:
                        error_text=traceback.format_exc()
                        add_log_info(current_datetime_stamp,"Gauge_Computation","error",error_text+"\n\nVacuum_QC_output: {}".format(Vacuum_QC_output))

# =============================================================================
#             else:
#                 logger.log('Vacuum Machine QC process initialized', 'success', False, None)
# =============================================================================
                    
            #... Overall QC Message To Be Published in UI ...#
            

            #.. If Both QC Processes Are Complete ...#
# =============================================================================
#             if stored_object['vacuum_qc']['qc_gauge_progress_percent']==100 and stored_object['vibration_qc']["qc_gauge_progress_percent"]==100:
#                 
#                 if stored_object['vacuum_qc']['qc_gauge_green_percent'] >= 30 and stored_object['vibration_qc']['qc_gauge_green_percent'] >= 30:
#                     logger.log('Overall Machine QC process completed', 'success', True, True)
#                     
#                 
#                 else:
#                     logger.log('Overall Machine QC process completed', 'success', True, False)
#                 
#                 if "qc_complete" not in stored_object or not stored_object["qc_complete"]:
#                     print("QC Complete")
#                     logger.update_qc_input_configuration(False)
#                     stored_object["qc_complete"]=True
#                     
#                     
#                     
#                 
#             
#             #.. Else, if Vacuum QC Process Is Complete But Vibration QC has not started yet, meaning the AC Sensor is disconnected ..#
#             elif stored_object['vacuum_qc']['qc_gauge_progress_percent']==100 and stored_object['vibration_qc']['qc_gauge_progress_percent']==0:
#                 
#                 if stored_object['vacuum_qc']['qc_gauge_green_percent'] >= 30:
#                     logger.log('Overall Machine QC process completed only with Vacuum', 'success', True, True)
#                 
#                 else:
#                     logger.log('Overall Machine QC process completed only with Vacuum', 'success', True, False)
#                 
#                 if "qc_complete" not in stored_object or not stored_object["qc_complete"]:
#                     print("QC Complete")
#                     logger.update_qc_input_configuration(False)
#                     stored_object["qc_complete"]=True
#             
#             #.. Else, if Vibration QC Process Is Complete But Vacuum QC has not started yet, meaning the VAC Sensor is disconnected ..#
#             elif stored_object['vacuum_qc']['qc_gauge_progress_percent']==0 and stored_object['vibration_qc']['qc_gauge_progress_percent']==100:
#                 
#                 if stored_object['vibration_qc']['qc_gauge_green_percent'] >= 30:
#                     logger.log('Overall Machine QC process completed only with Vibration', 'success', True, True)
#                 
#                 else:
#                     logger.log('Overall Machine QC process completed only with Vibration', 'success', True, False)
#                 
#                 if "qc_complete" not in stored_object or not stored_object["qc_complete"]:
#                     print("QC Complete")
#                     logger.update_qc_input_configuration(False)
#                     stored_object["qc_complete"]=True
#             
#             #.. If none of the above are satisfied, it means QC is still in progress
#             else:    
#                 logger.log('Overall Machine QC process completion pending', 'success', False, None)
# =============================================================================
                

            
        else:
            
            gauge_output = {}
            add_log_info(current_datetime_stamp,"Gauge_Computation","normal","manufacturer {} is not in the list of manufacturers in Gauge Computation".format(manufacturer))

    except Exception:
        
        error_text=traceback.format_exc()
        add_log_info(current_datetime_stamp,"Gauge_Computation","error",error_text)
        #traceback.print_exc()
        gauge_output={}

    
    return gauge_output
