"""
[-] Title: PUMP COMPUTATION
[~] Build: v2.0
[+] Author: Biprodeep Roy
[>] Licensed to: MachineSense™
[~] Description: This application retrives the raw data from critical process and processes it for pumpsense-analytics.
[~] Execute Command: python3 pump-computation.py -c /tmp/pumpsense-compute.json
"""
import time, os, glob, copy
import simplejson as json
import pump_analytics 
from BLEProcessor import BLEProcessor
from pumpsense_computation_lib.GracefulKiller import GracefullyKill
import threading
from pyanalytics_lib.LockingList import LockingList
from qc_computation.gauge_code.stored_objects_qc import read_stored_object

from sleeper import interruptable_sleep, awake_sleep
from pynovatec.events import EventPost,EventListener
from machineqc_events import MachineQCBatchCollectionComplete, MachineQCComplete, MachineQCShutdown, MachineQCInit, MachineQCError
from InputJsonReadyEvent import InputJsonReadyEvent
from pynovatec_compute.batchers import ZmqIntervalBatcher,CsvDataBatcher
from pynovatec_compute.batchers import DataCollectionEvents as DCEvents

csvFormat=1

config = {
            'CONFIG_FILE_PATH': '/var/r_config.json',
            'JSON_DATA_FOLDER_PATH': '/var/json_data',
            'CONF_FOLDER_PATH': '/usr/src/conf',
            'STORED_OBJECTS':'/home/pi/pumpsense_computation/stored_objects',
            'stored_objects_dir':'/home/pi/pumpsense_computation/stored_objects',
            'ANALYTICS_ERROR_LOG':'/home/pi/pumpsense_computation/reason_log.json',
            'INPUT_JSON_FOLDER_PATH':'/home/pi/pumpsense_computation/INPUT JSON',
            'OUTPUT_JSON_FOLDER_PATH':'/home/pi/pumpsense_computation/OUTPUT JSON',
            'bridge_data_path':'/home/pi/ramcsv/',
            'machine_filepath':'/var/json_data/machine_details_0.json',
            'send_input_json': False,
            'json_details_path': '/opt/novatec/slave/json_data',
            'working_dir': '/tmp/qc_computation',
            'log_file': '/tmp/qc/log/reason_log.json',
            'config_path': "/tmp/qc/conf/r_config.json",
            'conf_dir': "/tmp/qc/conf",
            'json_sender_port': "4445",
            'zmq_send_enabled': False,
            'json_send_output_json': False
}


# Temporary items for dealing with stored object.

def get_stored_object_file(machine_id):
    stored_object_path = os.path.join("/tmp/qc/stored_objects","stored_objects_" + machine_id + ".json")
    return stored_object_path

def get_stored_object(machine_id):
    so_file = get_stored_object_file(machine_id)
    return read_stored_object(so_file)

def remove_stored_object(machine_id):
    so_file = get_stored_object_file(machine_id)
    try:
        os.remove(so_file)
    except:
        pass

# End Temporary..


class MachineQC(EventListener):

    def __init__(self,ramcsv_path,machine_file_path,batcher="zmq",machines=[],interval=600):
        self.batcher = None

        print("Machine qc init...")
        self.ramcsv_path = ramcsv_path
        print("ramcsv_data_path: ",self.ramcsv_path)
        self.machine_filepath = machine_file_path
        self.pc = BLEProcessor(csv_format=csvFormat)
        self.gkiller = GracefullyKill()
        self.rl = LockingList()
        self.machines = machines
        self.machine_list = None
        batch_type = "test"
        #interval = 120
        #interval = 600
        super().__init__()
        self.events.append(DCEvents)
        self.register()
        self.keepRunning = True

        print("Batcher is: ",batcher) 
        if batcher == 'zmq':
            self.batcher = ZmqIntervalBatcher(interval=interval)
        elif batcher == 'csv':
            self.batcher = CsvDataBatcher(csv_path=ramcsv_path)

    def stop(self):
        print("Stopping...")
        self.keepRunning = False
        if self.batcher:
           self.batcher.stop()

        awake_sleep()

    def set_machine_list(self,machines):
        if not machines:
            return
        if self.machine_list is not None:
            return
        self.machine_list = {}
        templates = self.pc.get_templates()
        for machine_id in machines:
            if machine_id not in templates:
                EventPost(MachineQCError({"message": f"Machine ID ({machine_id}) not in json details"}))
                continue
            self.machine_list[machine_id] = {
                "machine_id": machine_id,
                "keep_running": True
            } 

    def cleanup(self):
        self.unregister()
        self.stop()  # Incase this isn't already stopped.
        if self.batcher:
            self.batcher.join()
            self.batcher = None

    def handle(self,msg):
        #print("Data Collection Listener run: ",msg.data)
        print("Data Collection Listener run: ",msg.event_name)
        #if msg.event_name == 'data_collection_event':
        if msg.event_name == 'raw_batch_data_event':
            if (self.keepRunning):
                jsons = self.pc.process_ble_list_tojson(msg.data['data'],timestamp=msg.data['timestamp'])
                EventPost(InputJsonReadyEvent(jsons,msg.data['duration']))
        if msg.event_name == 'input_json_ready':
            try:
                self.process_input_jsons(msg.data['data'])
            except Exception as e:
                print(e)

    def stored_object_cleanup(self):
        templates = self.pc.get_templates()
        for mid in templates:
            remove_stored_object(mid)

    def startup_events(self):
        #templates = self.pc.get_templates()
        all_templates = self.pc.get_templates()
        templates = {}
        for mid,item in self.machine_list.items():
            templates[mid] = all_templates[mid]
        data=[]
        for mid,item in templates.items():
            d={
                'machine_id': mid,
                'name': item['header']['machineName']
            }
            data.append(d)
        self.batcher.start()
        EventPost(MachineQCInit(data))

    def send_initial_data(self):
        #templates = self.pc.get_templates()
        all_templates = self.pc.get_templates()
        templates = {}
        for mid,item in self.machine_list.items():
            templates[mid] = all_templates[mid]
        print("send initial data templates: ",templates)
        for mid,item in templates.items():
            print("machine: ",item)
            if isinstance(item,dict):   
                   item_replicate = copy.deepcopy(item)
                   item_replicate['header']['timestamp'] = time.time()
                   del item_replicate['subassemblies']
                   print("Sending: ",json.dumps(item_replicate,indent=4))
                   pump_analytics.zmq_send_json("bogus",item_replicate)

    def extract_stored_object_qc(self,machine_id,qc_list):
        so = get_stored_object(machine_id)
        qc_so = {}
        for qc in qc_list:
           qc_so[qc] = so.get(qc,{})
        return qc_so

    def check_keep_running(self):
        if not self.machine_list:
            return False
        for machine_id,machine in self.machine_list.items():
            if machine['keep_running'] == True:
                return True
        return False

    def evaluate_gauge(self,result):
        gauges = result.get('gauge',{})
        count = 0
        machine_id = result['header']['machineId']
        if machine_id not in self.machine_list:
            return
        qc_so = self.extract_stored_object_qc(machine_id,["vibration_qc","vacuum_qc"])
        print("qc_so: ", json.dumps(qc_so,indent=4))
        keep_running = False
        ts = time.time()
        for qc_key,qc in qc_so.items():
            qc['timestamp'] = ts
            if qc.get('qc_gauge_progress_percent',100) != 100:
                keep_running =  True
        # This is bad, too specific.."
        if 'gauge' in result:
            if 'vacuumsense1:  Vacuum QC' in result['gauge']:
                if 'color' in result['gauge']['vacuumsense1:  Vacuum QC']:
                    qc_so['vacuum_qc']['color'] = result['gauge']['vacuumsense1:  Vacuum QC']['color'][0]
            if 'machinesense1:  Vibration QC' in result['gauge']:
                if 'color' in result['gauge']['machinesense1:  Vibration QC']:
                    qc_so['vibration_qc']['color'] = result['gauge']['machinesense1:  Vibration QC']['color'][0]

        self.machine_list[machine_id]['keep_running'] = keep_running
        qc_so['machine_id'] = machine_id
        print("QC POST: ",qc_so)
        EventPost(MachineQCBatchCollectionComplete(qc_so))
        if not keep_running:
            EventPost(MachineQCComplete(qc_so))
        if not self.check_keep_running():
            print("STOPPING..............")
            self.stop()

    def process_input_jsons(self,bmq):
        for macid in bmq:
            print("Main output macid: ",macid)
            #print(json.dumps(bmq[macid],indent=4))
            result = pump_analytics.main_output(bmq[macid])
            #print("Final Output: ",json.dumps(result,indent=4))
            pump_analytics.store_output_json(result)
            self.evaluate_gauge(result)
            print("After main output..")

    def main(self):
        #self.banner()
        self.initialise_input_json()

        if not self.machines:
            templates = self.pc.get_templates()
            self.machines = list(templates.keys())
        self.set_machine_list(self.machines)
        
        self.startup_events()
        self.send_initial_data()
        
        if not self.machine_list:
            EventPost(MachineQCError({"message": "No json details found for any machines"}))
            self.stop()
        self.stored_object_cleanup()
        while self.keepRunning:	# this is just a busy loop now, need to fix this.
            interruptable_sleep(60)
        self.cleanup()

    def initialise_input_json(self):
        mp = os.path.join(self.machine_filepath,'machine_details_*.json')
        machine_details_file_list_arr = glob.glob(mp)     
        for machine_details in machine_details_file_list_arr:
            self.pc.load_machine_file(machine_details)

class MachineQCThread(threading.Thread):
    def __init__(self,config,interval=60,batcher='zmq',machines=[]):
        super().__init__(target=self.run, name="MachineQCThread")
        pump_analytics.analytics_init(config)
        pump_analytics.zmq_init()
        batcher=config.get('batcher',batcher)
        self.pc = MachineQC(config['bridge_data_path'],config['machine_filepath'],interval=interval,batcher=batcher,machines=machines)

    def run(self):
        self.pc.main()
        print("Exiting...")
        EventPost(MachineQCShutdown({}))

    def stop(self):
        self.pc.stop()


